{% import 'includes/type_hint.py.jinja' as th -%}

# {% include 'includes/header.txt' %}

from collections.abc import Iterable
from typing import Union

import httpx
import httpx_auth
import lapidary.runtime.auth
from lapidary.runtime import NamedAuth
from typing_extensions import Literal
{% for auth_name, auth in model.security_schemes.items() %}

{% include "includes/security/" + auth.type + ".py.jinja" -%}
{% endfor -%}
