    def __init__(
        self,
        *,
        auth: {{ path.parent() }}.auth.Auth,
        base_url=
{%- if model.base_urls | length > 0 -%}
        '{{ model.base_urls[0] }}',
{%- else -%}
        None,
{%- endif %}
    ):
        ...
