from __future__ import annotations

import lapidary.runtime
import pydantic
import typing_extensions as typing

{%- for imp in item.imports %}
import {{ imp }}
{%- endfor %}
{% set path = item.path %}
{%- for model in item.body %}
{% include 'includes/schema/schema_class_' + model.model_type.name + '.py.jinja' %}
{%- endfor -%}
