# This file was generated but won't be updated automatically and may be edited manually.

[build-system]
build-backend = "poetry.core.masonry.api"
requires = ["poetry-core>=1.3.2"]

[tool.poetry]
name = "{{ config.package }}"
description = "Client library for {{ document.info.title }}"
version = "0.1.0"
authors = []
license = ""
packages = [
    { include = '{{ config.package }}', from = "gen" },
]

[tool.poetry.dependencies]
python = "^3.9"
lapidary = "^0.10.0"

[tool.poetry.group.dev.dependencies]
lapidary-render = "{{ get_version('lapidary_render') }}"

[tool.lapidary]
{%- for key, value in config.items() %}
{{ key }} = "{{ value }}"
{%- endfor %}
